/**************************************************************************************************
*                                                                                                 *
* This file is part of HPIPM.                                                                     *
*                                                                                                 *
* HPIPM -- High-Performance Interior Point Method.                                                *
* Copyright (C) 2019 by Gianluca Frison.                                                          *
* Developed at IMTEK (University of Freiburg) under the supervision of Moritz Diehl.              *
* All rights reserved.                                                                            *
*                                                                                                 *
* The 2-Clause BSD License                                                                        *
*                                                                                                 *
* Redistribution and use in source and binary forms, with or without                              *
* modification, are permitted provided that the following conditions are met:                     *
*                                                                                                 *
* 1. Redistributions of source code must retain the above copyright notice, this                  *
*    list of conditions and the following disclaimer.                                             *
* 2. Redistributions in binary form must reproduce the above copyright notice,                    *
*    this list of conditions and the following disclaimer in the documentation                    *
*    and/or other materials provided with the distribution.                                       *
*                                                                                                 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND                 *
* ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED                   *
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE                          *
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR                 *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES                  *
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;                    *
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND                     *
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT                      *
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS                   *
* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                    *
*                                                                                                 *
* Author: Gianluca Frison, gianluca.frison (at) imtek.uni-freiburg.de                             *
*                                                                                                 *
**************************************************************************************************/

/**
*
* Example of solving an OCP QP with data provided in the format generated by the routine
* d_ocp_qp_codegen(...);
* Additional features compared to the basic example:
* - initial value embedding (elimination of equality constraint on x0) before partial condensing
* - partial condensing of QP before calling the solver and partial expansion of QP solution after
*
**/

#include <stdlib.h>
#include <stdio.h>

#include <blasfeo_d_aux_ext_dep.h>

#include <hpipm_d_dense_qp_ipm.h>
#include <hpipm_d_dense_qp_dim.h>
#include <hpipm_d_dense_qp.h>
#include <hpipm_d_dense_qp_sol.h>
#include <hpipm_d_ocp_qp_dim.h>
#include <hpipm_d_ocp_qp.h>
#include <hpipm_d_ocp_qp_sol.h>
#include <hpipm_d_cond.h>
#include <hpipm_d_ocp_qp_red.h>
#include <hpipm_d_ocp_qp_utils.h>
#include <hpipm_timing.h>



// qp data as global data
extern int N;
extern int *nx;
extern int *nu;
extern int *nbu;
extern int *nbx;
extern int *ng;
extern int *nsbx;
extern int *nsbu;
extern int *nsg;
extern int *nbue;
extern int *nbxe;
extern int *nge;
extern double **hA;
extern double **hB;
extern double **hb;
extern double **hQ;
extern double **hR;
extern double **hS;
extern double **hq;
extern double **hr;
extern int **hidxbx;
extern double **hlbx;
extern double **hlbx_mask;
extern double **hubx;
extern double **hubx_mask;
extern int **hidxbu;
extern double **hlbu;
extern double **hlbu_mask;
extern double **hubu;
extern double **hubu_mask;
extern double **hC;
extern double **hD;
extern double **hlg;
extern double **hlg_mask;
extern double **hug;
extern double **hug_mask;
extern double **hZl;
extern double **hZu;
extern double **hzl;
extern double **hzu;
extern int **hidxs;
extern double **hlls;
extern double **hlus;
// TODO mask soft constraints
extern int **hidxe;
// arg
extern int mode;
extern int iter_max;
extern double alpha_min;
extern double mu0;
extern double tol_stat;
extern double tol_eq;
extern double tol_ineq;
extern double tol_comp;
extern double reg_prim;
extern int warm_start;
extern int pred_corr;
extern int ric_alg;
extern int split_step;



// main
int main()
	{

	int ii;

	int hpipm_status;

	int rep, nrep=10;

	hpipm_timer timer;

/************************************************
* ocp qp dim
************************************************/

	hpipm_size_t dim_size = d_ocp_qp_dim_memsize(N);
	void *dim_mem = malloc(dim_size);

	struct d_ocp_qp_dim dim;
	d_ocp_qp_dim_create(N, &dim, dim_mem);

	d_ocp_qp_dim_set_all(nx, nu, nbx, nbu, ng, nsbx, nsbu, nsg, &dim);

	// set number of inequality constr to be considered as equality constr
//	d_ocp_qp_dim_set_nbxe(0, nx[0], &dim);
	for(ii=0; ii<=N; ii++)
		{
		d_ocp_qp_dim_set_nbxe(ii, nbxe[ii], &dim);
		d_ocp_qp_dim_set_nbue(ii, nbue[ii], &dim);
		d_ocp_qp_dim_set_nge(ii, nge[ii], &dim);
		}

//	d_ocp_qp_dim_codegen("examples/c/data/test_d_ocp_data.c", "w", &dim);

/************************************************
* ocp qp dim red eq dof (reduce equation dof, i.e. x0 elimination)
************************************************/

	hpipm_size_t dim_size3 = d_ocp_qp_dim_memsize(N);
	void *dim_mem3 = malloc(dim_size3);

	struct d_ocp_qp_dim dim3;
	d_ocp_qp_dim_create(N, &dim3, dim_mem3);

	d_ocp_qp_dim_reduce_eq_dof(&dim, &dim3);

/************************************************
* dense qp dim cond
************************************************/

	hpipm_size_t dim_size2 = d_dense_qp_dim_memsize();
	void *dim_mem2 = malloc(dim_size2);

	struct d_dense_qp_dim dim2;
	d_dense_qp_dim_create(&dim2, dim_mem2);

	d_cond_qp_compute_dim(&dim3, &dim2);

/************************************************
* ocp qp
************************************************/

	hpipm_size_t qp_size = d_ocp_qp_memsize(&dim);
	void *qp_mem = malloc(qp_size);

	struct d_ocp_qp qp;
	d_ocp_qp_create(&dim, &qp, qp_mem);

	d_ocp_qp_set_all(hA, hB, hb, hQ, hS, hR, hq, hr, hidxbx, hlbx, hubx, hidxbu, hlbu, hubu, hC, hD, hlg, hug, hZl, hZu, hzl, hzu, hidxs, hlls, hlus, &qp);

	// mark the inequality constr on x0 as an equality constr
//	int *idxbxe0 = (int *) malloc(nx[0]*sizeof(int));
//	for(ii=0; ii<=nx[0]; ii++)
//		idxbxe0[ii] = ii;
//	d_ocp_qp_set_idxbxe(0, idxbxe0, &qp);
	for(ii=0; ii<=N; ii++)
		{
		d_ocp_qp_set_idxe(ii, hidxe[ii], &qp);
		}

	// set inequality constraints mask
	for(ii=0; ii<=N; ii++)
		{
		d_ocp_qp_set_lbu_mask(ii, hlbu_mask[ii], &qp);
		d_ocp_qp_set_ubu_mask(ii, hubu_mask[ii], &qp);
		d_ocp_qp_set_lbx_mask(ii, hlbx_mask[ii], &qp);
		d_ocp_qp_set_ubx_mask(ii, hubx_mask[ii], &qp);
		d_ocp_qp_set_lg_mask(ii, hlg_mask[ii], &qp);
		d_ocp_qp_set_ug_mask(ii, hug_mask[ii], &qp);
		}

//	d_ocp_qp_codegen("examples/c/data/test_d_ocp_data.c", "a", &dim, &qp);

/************************************************
* ocp qp red eq dof
************************************************/

	hpipm_size_t qp_size3 = d_ocp_qp_memsize(&dim3);
	void *qp_mem3 = malloc(qp_size3);

	struct d_ocp_qp qp3;
	d_ocp_qp_create(&dim3, &qp3, qp_mem3);

/************************************************
* dense qp cond
************************************************/

	hpipm_size_t qp_size2 = d_dense_qp_memsize(&dim2);
	void *qp_mem2 = malloc(qp_size2);

	struct d_dense_qp qp2;
	d_dense_qp_create(&dim2, &qp2, qp_mem2);

/************************************************
* ocp qp sol
************************************************/

	hpipm_size_t qp_sol_size = d_ocp_qp_sol_memsize(&dim);
	void *qp_sol_mem = malloc(qp_sol_size);

	struct d_ocp_qp_sol qp_sol;
	d_ocp_qp_sol_create(&dim, &qp_sol, qp_sol_mem);

/************************************************
* ocp qp sol red eq dof
************************************************/

	hpipm_size_t qp_sol_size3 = d_ocp_qp_sol_memsize(&dim3);
	void *qp_sol_mem3 = malloc(qp_sol_size3);

	struct d_ocp_qp_sol qp_sol3;
	d_ocp_qp_sol_create(&dim3, &qp_sol3, qp_sol_mem3);

/************************************************
* dense qp sol cond
************************************************/

	hpipm_size_t qp_sol_size2 = d_dense_qp_sol_memsize(&dim2);
	void *qp_sol_mem2 = malloc(qp_sol_size2);

	struct d_dense_qp_sol qp_sol2;
	d_dense_qp_sol_create(&dim2, &qp_sol2, qp_sol_mem2);

/************************************************
* red eq dof arg
************************************************/

	hpipm_size_t qp_red_arg_size = d_ocp_qp_reduce_eq_dof_arg_memsize();
	void *qp_red_arg_mem = malloc(qp_red_arg_size);

	struct d_ocp_qp_reduce_eq_dof_arg qp_red_arg;
	d_ocp_qp_reduce_eq_dof_arg_create(&qp_red_arg, qp_red_arg_mem);

	int tmp_i1 = 1;
	d_ocp_qp_reduce_eq_dof_arg_set_default(&qp_red_arg);
	d_ocp_qp_reduce_eq_dof_arg_set_alias_unchanged(&tmp_i1, &qp_red_arg);
	d_ocp_qp_reduce_eq_dof_arg_set_comp_dual_sol_eq(&tmp_i1, &qp_red_arg);
	d_ocp_qp_reduce_eq_dof_arg_set_comp_dual_sol_ineq(&tmp_i1, &qp_red_arg);

/************************************************
* cond arg
************************************************/

	hpipm_size_t cond_arg_size = d_cond_qp_arg_memsize();
	void *cond_arg_mem = malloc(cond_arg_size);

	struct d_cond_qp_arg cond_arg;
	d_cond_qp_arg_create(&cond_arg, cond_arg_mem);

	d_cond_qp_arg_set_default(&cond_arg);

//	d_cond_qp_arg_set_ric_alg(0, &cond_arg);

/************************************************
* ipm arg
************************************************/

	hpipm_size_t ipm_arg_size = d_dense_qp_ipm_arg_memsize(&dim2);
	void *ipm_arg_mem = malloc(ipm_arg_size);

	struct d_dense_qp_ipm_arg arg;
	d_dense_qp_ipm_arg_create(&dim2, &arg, ipm_arg_mem);

	d_dense_qp_ipm_arg_set_default(mode, &arg);

	d_dense_qp_ipm_arg_set_mu0(&mu0, &arg);
	d_dense_qp_ipm_arg_set_iter_max(&iter_max, &arg);
	d_dense_qp_ipm_arg_set_alpha_min(&alpha_min, &arg);
	d_dense_qp_ipm_arg_set_mu0(&mu0, &arg);
	d_dense_qp_ipm_arg_set_tol_stat(&tol_stat, &arg);
	d_dense_qp_ipm_arg_set_tol_eq(&tol_eq, &arg);
	d_dense_qp_ipm_arg_set_tol_ineq(&tol_ineq, &arg);
	d_dense_qp_ipm_arg_set_tol_comp(&tol_comp, &arg);
	d_dense_qp_ipm_arg_set_reg_prim(&reg_prim, &arg);
	d_dense_qp_ipm_arg_set_warm_start(&warm_start, &arg);
	d_dense_qp_ipm_arg_set_pred_corr(&pred_corr, &arg);
	d_dense_qp_ipm_arg_set_split_step(&split_step, &arg);

//	d_dense_qp_ipm_arg_codegen("examples/c/data/test_d_ocp_data.c", "a", &dim2, &arg);

/************************************************
* red eq dof workspace
************************************************/

	hpipm_size_t qp_red_work_size = d_ocp_qp_reduce_eq_dof_ws_memsize(&dim);
	void *qp_red_work_mem = malloc(qp_red_work_size);

	struct d_ocp_qp_reduce_eq_dof_ws qp_red_work;
	d_ocp_qp_reduce_eq_dof_ws_create(&dim, &qp_red_work, qp_red_work_mem);

/************************************************
* cond workspace
************************************************/

	hpipm_size_t cond_size = d_cond_qp_ws_memsize(&dim3, &cond_arg);
	void *cond_mem = malloc(cond_size);

	struct d_cond_qp_ws cond_ws;
	d_cond_qp_ws_create(&dim3, &cond_arg, &cond_ws, cond_mem);

/************************************************
* ipm workspace
************************************************/

	hpipm_size_t ipm_size = d_dense_qp_ipm_ws_memsize(&dim2, &arg);
	void *ipm_mem = malloc(ipm_size);

	struct d_dense_qp_ipm_ws workspace;
	d_dense_qp_ipm_ws_create(&dim2, &arg, &workspace, ipm_mem);

/************************************************
* reduce equation dof (i.e. x0 elimination)
************************************************/

	hpipm_tic(&timer);

	for(rep=0; rep<nrep; rep++)
		{
		d_ocp_qp_reduce_eq_dof(&qp, &qp3, &qp_red_arg, &qp_red_work);
		}

	double time_red_eq_dof = hpipm_toc(&timer) / nrep;

/************************************************
* part cond
************************************************/

	hpipm_tic(&timer);

	for(rep=0; rep<nrep; rep++)
		{
		d_cond_qp_cond(&qp3, &qp2, &cond_arg, &cond_ws);
		}

	double time_cond = hpipm_toc(&timer) / nrep;

	//d_dense_qp_codegen_matlab("examples/c/data/ocp_data.m", "w", qp2.dim, &qp2);

/************************************************
* ipm solver
************************************************/

	hpipm_tic(&timer);

	for(rep=0; rep<nrep; rep++)
		{
		d_dense_qp_ipm_solve(&qp2, &qp_sol2, &arg, &workspace);
		d_dense_qp_ipm_get_status(&workspace, &hpipm_status);
		}

	double time_ipm = hpipm_toc(&timer) / nrep;

/************************************************
* part expand
************************************************/

	hpipm_tic(&timer);

	for(rep=0; rep<nrep; rep++)
		{
		d_cond_qp_expand_sol(&qp3, &qp_sol2, &qp_sol3, &cond_arg, &cond_ws);
		}

	double time_expa = hpipm_toc(&timer) / nrep;

/************************************************
* restore equation dof
************************************************/

	hpipm_tic(&timer);

	for(rep=0; rep<nrep; rep++)
		{
		d_ocp_qp_restore_eq_dof(&qp, &qp_sol3, &qp_sol, &qp_red_arg, &qp_red_work);
		}

	double time_res_eq_dof = hpipm_toc(&timer) / nrep;

/************************************************
* print solution info
************************************************/

    printf("\nHPIPM returned with flag %i.\n", hpipm_status);
    if(hpipm_status == 0)
		{
        printf("\n -> QP solved!\n");
		}
	else if(hpipm_status==1)
		{
        printf("\n -> Solver failed! Maximum number of iterations reached\n");
		}
	else if(hpipm_status==2)
		{
        printf("\n -> Solver failed! Minimum step lenght reached\n");
		}
	else if(hpipm_status==3)
		{
        printf("\n -> Solver failed! NaN in computations\n");
		}
	else
		{
        printf("\n -> Solver failed! Unknown return flag\n");
		}
    printf("\nAverage solution time over %i runs: %e [s]\n", nrep, time_ipm);
	printf("\n\n");

/************************************************
* extract and print solution
************************************************/

	// u

	int nu_max = nu[0];
	for(ii=1; ii<=N; ii++)
		if(nu[ii]>nu_max)
			nu_max = nu[ii];

	double *u = malloc(nu_max*sizeof(double));

	printf("\nu = \n");
	for(ii=0; ii<=N; ii++)
		{
		d_ocp_qp_sol_get_u(ii, &qp_sol, u);
		d_print_mat(1, nu[ii], u, 1);
		}

	// x

	int nx_max = nx[0];
	for(ii=1; ii<=N; ii++)
		if(nx[ii]>nx_max)
			nx_max = nx[ii];

	double *x = malloc(nx_max*sizeof(double));

	printf("\nx = \n");
	for(ii=0; ii<=N; ii++)
		{
		d_ocp_qp_sol_get_x(ii, &qp_sol, x);
		d_print_mat(1, nx[ii], x, 1);
		}

//	d_ocp_qp_sol_print(&dim, &qp_sol);

/************************************************
* print ipm statistics
************************************************/

	int iter; d_dense_qp_ipm_get_iter(&workspace, &iter);
	double res_stat; d_dense_qp_ipm_get_max_res_stat(&workspace, &res_stat);
	double res_eq; d_dense_qp_ipm_get_max_res_eq(&workspace, &res_eq);
	double res_ineq; d_dense_qp_ipm_get_max_res_ineq(&workspace, &res_ineq);
	double res_comp; d_dense_qp_ipm_get_max_res_comp(&workspace, &res_comp);
	double *stat; d_dense_qp_ipm_get_stat(&workspace, &stat);
	int stat_m; d_dense_qp_ipm_get_stat_m(&workspace, &stat_m);

	printf("\nipm return = %d\n", hpipm_status);
	printf("\nipm residuals max: res_g = %e, res_b = %e, res_d = %e, res_m = %e\n", res_stat, res_eq, res_ineq, res_comp);

	printf("\nipm iter = %d\n", iter);
	printf("\nalpha_aff\tmu_aff\t\tsigma\t\talpha_prim\talpha_dual\tmu\t\tres_stat\tres_eq\t\tres_ineq\tres_comp\tobj\t\tlq fact\t\titref pred\titref corr\tlin res stat\tlin res eq\tlin res ineq\tlin res comp\n");
	d_print_exp_tran_mat(stat_m, iter+1, stat, stat_m);

	printf("\neq dof red time = %e [s]\n\n", time_red_eq_dof);
	printf("\npart cond time = %e [s]\n\n", time_cond);
	printf("\nocp ipm time = %e [s]\n\n", time_ipm);
	printf("\npart expa time = %e [s]\n\n", time_expa);
	printf("\neq dof res time = %e [s]\n\n", time_res_eq_dof);
	printf("\ntotal time = %e [s]\n\n", time_red_eq_dof+time_cond+time_ipm+time_expa+time_res_eq_dof);

/************************************************
* free memory and return
************************************************/

	free(dim_mem);
	free(dim_mem2);
	free(dim_mem3);
	free(qp_mem);
	free(qp_mem2);
	free(qp_mem3);
	free(qp_sol_mem);
	free(qp_sol_mem2);
	free(qp_sol_mem3);
//	free(idxbxe0);
    free(qp_red_arg_mem);
    free(qp_red_work_mem);
	free(cond_arg_mem);
	free(ipm_arg_mem);
	free(cond_mem);
	free(ipm_mem);

	free(u);
	free(x);

	return 0;

	}




